/*
 * This test program largely adapted from the android SDK camera preview
 * example which has this copyright:
 *
 * Copyright (C) 2007 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package lan.my.zooty.andyscan;

import android.content.Context;
import android.hardware.Camera;
import android.hardware.Camera.Size;
import android.hardware.Camera.PreviewCallback;
import android.os.Bundle;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.view.ViewGroup;
import android.util.AttributeSet;
import android.widget.TextView;

import java.io.IOException;
import java.util.List;

/**
 * A simple wrapper around a Camera and a SurfaceView that renders a
 * centered preview of the Camera to the surface. We need to center the
 * SurfaceView because not all devices have cameras that support preview
 * sizes at the same aspect ratio as the device's display.
 */
class Preview extends ViewGroup implements SurfaceHolder.Callback {
    private final String TAG = "Preview";

    SurfaceView mSurfaceView;
    SurfaceHolder mHolder;
    Size mPreviewSize;
    List<Size> mSupportedPreviewSizes;
    Camera mCamera;
    byte[] lastBuffer = null;
    TextView lightText;
    TextView evText;
    TextView motionText;

    // This custom PreviewCallback adds up the Y part of the YUV
    // preview image to get an idea of how bright things are
    // in front of the camera (this is one idea for a possible
    // trigger for taking a new picture). It runs as a one shot
    // callback so the preview data won't be disturbed while
    // I'm fooling with it, then it triggers the next one shot.
    // Not going to be a high frame rate, but that shouldn't
    // be needed.

    PreviewCallback mPreviewCallback = new PreviewCallback() {
       @Override
       public void onPreviewFrame(byte[] data, Camera camera) {
          int byteCount = mPreviewSize.width *
                          mPreviewSize.height;
          if (lightText != null) {
             int luma = 0;
             for (int i = 0; i < byteCount; ++i) {
                luma += (((int)(data[i])) & 0xff);
             }
             String lumastr = "" + luma;
             lightText.setText((CharSequence)lumastr);
          }
          if (evText != null) {
             Camera.Parameters p = mCamera.getParameters();
             int evcomp = p.getExposureCompensation();
             float evstep = p.getExposureCompensationStep();
             int minev = p.getMinExposureCompensation();
             int maxev = p.getMaxExposureCompensation();
             if (minev != maxev) {
                evstep = evstep * (float)evcomp;
                String evstr = Float.toString(evstep);
                evText.setText((CharSequence)evstr);
             }
          }
          if ((lastBuffer != null) && (motionText != null)) {
             int diff = 0;
             for (int j = 0; j < byteCount; ++j) {
                int d = (((int)(data[j])) & 0xff) -
                        (((int)(lastBuffer[j])) & 0xff);
                if (d < 0) {
                   d = - d;
                }
                diff += d;
             }
             String motionstr = "" + diff;
             motionText.setText((CharSequence)motionstr);
             mCamera.addCallbackBuffer(lastBuffer);
          }
          lastBuffer = data;
       }
    };

    public
    Preview(Context context, AttributeSet attrs) {
        super(context, attrs);

        mSurfaceView = new SurfaceView(context);
        addView(mSurfaceView);

        // Install a SurfaceHolder.Callback so we get notified when the
        // underlying surface is created and destroyed.
        mHolder = mSurfaceView.getHolder();
        mHolder.addCallback(this);
        mHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);
    }

    public void setCamera(Camera camera) {
        mCamera = camera;
        if (mCamera != null) {
            mSupportedPreviewSizes = mCamera.getParameters().getSupportedPreviewSizes();
            requestLayout();
        }
    }

    public void setLightText(TextView newText, TextView newEv) {
        lightText = newText;
        evText = newEv;
    }

    public void setMotionText(TextView newText) {
        motionText = newText;
    }

    @Override
    protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        // We purposely disregard child measurements because act as a
        // wrapper to a SurfaceView that centers the camera preview instead
        // of stretching it.
        final int width = resolveSize(getSuggestedMinimumWidth(), widthMeasureSpec);
        final int height = resolveSize(getSuggestedMinimumHeight(), heightMeasureSpec);
        setMeasuredDimension(width, height);

        if (mSupportedPreviewSizes != null) {
            mPreviewSize = getOptimalPreviewSize(mSupportedPreviewSizes, width, height);
        }
    }

    @Override
    protected void onLayout(boolean changed, int l, int t, int r, int b) {
        if (changed && getChildCount() > 0) {
            final View child = getChildAt(0);

            final int width = r - l;
            final int height = b - t;

            int previewWidth = width;
            int previewHeight = height;
            if (mPreviewSize != null) {
                previewWidth = mPreviewSize.width;
                previewHeight = mPreviewSize.height;
            }

            // Center the child SurfaceView within the parent.
            if (width * previewHeight > height * previewWidth) {
                final int scaledChildWidth = previewWidth * height / previewHeight;
                child.layout((width - scaledChildWidth) / 2, 0,
                        (width + scaledChildWidth) / 2, height);
            } else {
                final int scaledChildHeight = previewHeight * width / previewWidth;
                child.layout(0, (height - scaledChildHeight) / 2,
                        width, (height + scaledChildHeight) / 2);
            }
        }
    }

    private Size getOptimalPreviewSize(List<Size> sizes, int w, int h) {
        final double ASPECT_TOLERANCE = 0.1;
        double targetRatio = (double) w / h;
        if (sizes == null) return null;

        Size optimalSize = null;
        double minDiff = Double.MAX_VALUE;

        int targetHeight = h;

        // Try to find an size match aspect ratio and size
        for (Size size : sizes) {
            double ratio = (double) size.width / size.height;
            if (Math.abs(ratio - targetRatio) > ASPECT_TOLERANCE) continue;
            if (Math.abs(size.height - targetHeight) < minDiff) {
                optimalSize = size;
                minDiff = Math.abs(size.height - targetHeight);
            }
        }

        // Cannot find the one match the aspect ratio, ignore the requirement
        if (optimalSize == null) {
            minDiff = Double.MAX_VALUE;
            for (Size size : sizes) {
                if (Math.abs(size.height - targetHeight) < minDiff) {
                    optimalSize = size;
                    minDiff = Math.abs(size.height - targetHeight);
                }
            }
        }
        return optimalSize;
    }

    public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
        // Now that the size is known, set up the camera parameters and begin
        // the preview.
        Camera.Parameters parameters = mCamera.getParameters();
        parameters.setPreviewSize(mPreviewSize.width, mPreviewSize.height);
        requestLayout();

        // These don't exist in the old version of android on my phone :-(.
        //
        // if (parameters.isAutoExposureLockSupported()) {
        //    parameters.setAutoExposureLock(true);
        // }
        // if (parameters.isAutoWhiteBalanceLockSupported()) {
        //    parameters.setAutoWhiteBalanceLock(true);
        // }

        mCamera.setParameters(parameters);
        mCamera.startPreview();
        mCamera.addCallbackBuffer(
           new byte[2 * mPreviewSize.width * mPreviewSize.height]);
        mCamera.addCallbackBuffer(
           new byte[2 * mPreviewSize.width * mPreviewSize.height]);
        mCamera.setPreviewCallbackWithBuffer(mPreviewCallback);
    }

    public void surfaceCreated(SurfaceHolder holder) {
        // The Surface has been created, acquire the camera and tell it where
        // to draw.
        try {
            if (mCamera != null) {
                mCamera.setPreviewDisplay(holder);
            }
        } catch (IOException exception) {
            Log.e(TAG, "IOException caused by setPreviewDisplay()", exception);
        }
    }

    public void surfaceDestroyed(SurfaceHolder holder) {
        // Surface will be destroyed when we return, so stop the preview.
        if (mCamera != null) {
            mCamera.stopPreview();
        }
    }

}
